/* capture.c (c) 2004 Petr Danecek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define _GNU_SOURCE
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <getopt.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "capture.h"
#include "version.h"

char *SOCK_PATH = "/tmp/.capture_socket";
FILE *log_file  = NULL;

int msg_get(char *str, int len)
{
      int n;
      n = recv(SOCKET, str, 1, 0);
       
      if ( n <= 0 ) 
      {
	    if ( errno == EAGAIN ) return 0;

	    debug("msg_get %d: error %d - ", SOCKET, errno);
	    error("get: %s\n", strerror(errno));
	    exit(1);
      }
     
      n = (int) *str;
      if ( n > len ) 
      {
	    error("msg_recv FIXME: len (%d) < %d!\n", len,n);
	    n = len - 1;
      }

      n = recv(SOCKET, str, n, 0);
      if ( n <= 0)
      {
	    debug("msg_get2 %d: error %d - ", SOCKET, errno);
	    error("get: %s\n", strerror(errno));
	    exit(1);
      }
     
      str[n] = 0;
      debug("msg_get %d: %s\n", SOCKET, str);
      return n;
}

int msg_send(const char *format, ...)
{
      unsigned char buf[MAX_CMD_LEN];
      int n;
      
      va_list ap;
      va_start(ap, format);
      vsnprintf((char*)buf+1, MAX_CMD_LEN-1, format, ap);
      va_end(ap);
      
      n = strlen((char*)buf+1);
      *buf  = (unsigned char) n;

      if ( n <= 0 ) error("FIXME: empty msg\n");
      if ( (int)*buf != n ) error("msg_send FIXME: MAX_CMD_LEN exceeds 1byte: %d!=%d\n", (int)*buf,n);
     
      debug("msg_send %d: %s\n", SOCKET, buf+1);
      if ( send(SOCKET, buf, n+1, 0) == -1 ) 
      {
	    error("send: %s\n", strerror(errno));
	    exit(1);
      }
      return n;
}


void msg_cmd_ok(char *cmd)
{
      msg_send("%s %s", cmd, "ok");
}


void msg_cmd_info(char *cmd, const char *format, ...)
{
      char buf[MAX_CMD_LEN];
      
      va_list ap;
      va_start(ap, format);
      vsnprintf(buf, MAX_CMD_LEN, format, ap);
      va_end(ap);
      
      msg_send("%s info: %s", cmd, buf);
}


void msg_cmd_error(char *cmd, const char *format, ...)
{
      char buf[MAX_CMD_LEN];
      
      va_list ap;
      va_start(ap, format);
      vsnprintf(buf, MAX_CMD_LEN, format, ap);
      va_end(ap);
      
      msg_send("%s error: %s", cmd, buf);
}


void debug(const char *format, ...)
{
      va_list ap;
      va_start(ap, format);

      if ( log_file )
      {
	    flockfile(log_file);
	    vfprintf(log_file, format, ap);
	    fflush(log_file);
	    funlockfile(log_file);
      }
      va_end(ap);

      if ( ! DEBUG ) return;

      vfprintf(stderr, format, ap);
}


void error(const char *format, ...)
{
      va_list ap;
      va_start(ap, format);
      vfprintf(stderr, format, ap);
      va_end(ap);

      if ( log_file )
      {
	    flockfile(log_file);
	    vfprintf(log_file, format, ap);
	    fflush(log_file);
	    funlockfile(log_file);
      }
}


void print(const char *format, ...)
{
      va_list ap;
      va_start(ap, format);
      vprintf(format, ap);
      va_end(ap);

      if ( log_file )
      {
	    flockfile(log_file);
	    vfprintf(log_file, format, ap);
	    fflush(log_file);
	    funlockfile(log_file);
      }
}


void main_help()
{
      printf("Usage: capture [OPTIONS] command1 command2 ... \n\n"
	    "Command line options:\n");
      printf(
	    "     -h, --help                 this help message\n"
	    "     -s, --server               force server to run in foreground\n"
	    "     -S, --socket <file>        use this file for connection (default is \"%s\")\n"
#ifdef HAVE_GTK
	    "     -v, --viewfinder           viewfinder window\n"
#endif
	    "     -l, --log <file>           log all messages to a file\n"
	    "     -d, --debug                print lots of debug messages\n"
	    "     -V, --version              print version information and exit\n"
	    "\n"
	    "Report bugs to https://lists.sourceforge.net/lists/listinfo/capture-devel\n"
	    "\n", SOCK_PATH
      );

      help(NULL);
}


void print_version()
{
      printf("capture %s\n\n"
	    "Copyright (C) 2002 Free Software Foundation, Inc.\n"
	    "This is free software; see the source for copying conditions.  There is NO\n"
	    "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
	    , CAPTURE_VERSION);
}


void help(char *cmd)
{
      allowed_t *all;
      cmd_t *c;
      if ( cmd )
      {
	    int n;
	    cmd = getarg(cmd, &n);
	    all = NULL;
	    if ( cmd )
	    {
		  all = get_allowed_values(cmd+n);
		  while (all && all->string)
		  {
			printf("\t%s\n", all->string);
			all++;
		  }
	    }

	    if ( all ) return;
      }
      
      printf("Interactive commands:\n");
      c = cmds;
      while ( c->command )
      {
	    printf("\t%-20s%s\n", c->command, (c->help) ? c->help : "");
	    c++;
      }
}


int main(int argc, char **argv)
{
      int len;
      struct sockaddr_un remote;
      int option_index = 0, opt;
      struct stat sbuf;
      static struct option loptions[] = 
      {
	    {"help",0,0,'h'},
	    {"server",0,0,'s'},
	    {"socket",0,0,'S'},
#ifdef HAVE_GTK
	    {"viewfinder",0,0,'v'},
#endif
	    {"log",1,0,'l'},
	    {"debug",0,0,'d'},
	    {"version",0,0,'V'},
	    {0,0,0,0}
      };

      int server               = 0;

#ifdef HAVE_GTK
      settings.gtk             = 0;        /* is there some GTK stuff present ? */
      settings.viewfinder      = 0;        /* is there a viewfinder enabled ? */
      settings.bordersview     = 0;        /* no borders at start (percent of the width/height) */
      settings.alignview       = 0;        /* vfalign mode (see properties.c -> vfalign_vals) */
      settings.alignfname      = NULL;     /* filename of the vfalign'ed image, malloc() a string with
					      filename to update the image vfalign'ed. */
      settings.tmpalign        = NULL;     /* viewfinder snapshot */
      settings.viewfinderwait  = 0;        /* used to suspend viewfinder refreshes while capturing */
      settings.viewfinder_opacity = 70;    /* default: 70% opacity */
      settings.vfsave_file     = NULL;     /* default is not to save vf images */
#endif

      while (1)
      {
	    opt = getopt_long (argc, argv, "hsdvl:S:V", loptions, &option_index);
	    if ( opt == -1 ) break;

	    switch (opt)
	    {
		  case 's': server  = 1; break;
		  case 'S': SOCK_PATH = optarg; break;
		  case 'd': DEBUG = 1; break;
#ifdef HAVE_GTK
		  case 'v': settings.viewfinder = 1; break;
#endif
		  case 'l': 
			log_file = fopen(optarg,"a+"); 
			if ( ! log_file ) { error("fopen: %s\n", strerror(errno)); return 0; }
			break;
			
		  case 'V':
			print_version(); return 0;
		  
		  case 'h': 
		  default:
			main_help(); return 0;
	    }
      }

      if ( server )
      {
	    debug("Running in server mode using socket \"%s\".\n", SOCK_PATH);
	    run_server();
	    return 0;
      }

      remote.sun_family = AF_UNIX;
      strcpy(remote.sun_path, SOCK_PATH);
      len = strlen(remote.sun_path) + sizeof(remote.sun_family);

      if ((SOCKET = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) 
      {
	    error("socket: %s\n", strerror(errno));
	    return 0;
      }

      if ( stat(SOCK_PATH, &sbuf) == 0 )
      {
	    debug("Trying to connect to \"%s\" ...\n", SOCK_PATH);
    
	    if ( connect(SOCKET, (struct sockaddr *)&remote, len) != -1) 
	    {
		  return run_client(argc-optind, argv+optind);
	    }
	    
	    debug("Could not connect - fork\n");
	    unlink(remote.sun_path);
      }
      
      
      if ( fork() )
      {
	    usleep(USLEEP);
	    debug("Trying to connect to \"%s\" ...\n", SOCK_PATH);

	    if ( connect(SOCKET, (struct sockaddr *)&remote, len) == -1)
	    {
		  error("connect: %s\n", strerror(errno));
		  return 1;
	    }
	    return run_client(argc-optind, argv+optind);
      }
      else
      {
	    debug("Running server using socket \"%s\".\n", SOCK_PATH);
	    run_server();
	    return 0;
      }

      return 0;
}



