/* capture.c (c) 2004 Petr Danecek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#ifdef HAVE_READLINE
  #include <readline/readline.h>
  #include <readline/history.h>
#endif
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <signal.h>
#include "capture.h"



/*  The client sent "command [parameters]". Now we check the response of the server:
    Either "command ok", "command info: message" or "command error[: message]".
    Returns 0 on error, 1 on ok and 2 on info.
*/
int cmd_ok(char *cmd, char *ret)
{
      int n1, n2;
      char *arg1 = getarg(cmd, &n1);
      char *arg2 = getarg(ret, &n2);

      if ( n1 != n2 ) return 0;
      if ( strncasecmp(arg1, arg2, n1) ) return 0;

      arg2 = getarg(arg2+n2, &n2);
      if ( n2==2 && !strncasecmp("ok", arg2, n2) ) return 1;
      if ( n2==5 && !strncasecmp("info:", arg2, n2) ) return 2;

      return 0;
}


/* Run the given command. If the returned message is not "command ok",
    the message will be printed. See capture.h for currently recognised
    variants. The method returns 0 on success, -1 on non-fatal error and
    1 when the client should exit.
*/
int run_command(char *str)
{
    int status;
    char buf[MAX_CMD_LEN];
    if ( ! cmdcmp(str, CMD_HELP) ) 
    {
        help(str);
        return 0;
    }

    msg_send(str);

    /* wait for reply */
    while ( ! msg_get(buf, MAX_CMD_LEN) ) usleep(USLEEP);

    status = cmd_ok(str, buf);

    if ( status==0 ) 
    {
        error("%s\n", buf);
        return -1;
    }
    if ( status==2 )
        printf("%s\n", buf);

    if ( ! cmdcmp(buf, CMD_QUIT) ) return 1;
    if ( ! cmdcmp(buf, CMD_LEAVE) ) return 1;

    if ( ! cmdcmp(buf, CMD_STATUS) ) printf("%s\n", buf);
    else if ( ! cmdcmp(buf, CMD_UNKNOWN) ) help(NULL);

    return 0;
}


void sigterm_handler(int signal, struct sigcontext sc)
{
      debug("client signal caught: %d\n", signal);
      run_command(CMD_QUIT);
      exit(1);
}

int run_client(int argc, char *argv[])
{
    int stat;
    char *s;
    char str[MAX_CMD_LEN];
    int i;

    debug("Client: Connected \"%s\" as %d.\n", SOCK_PATH, SOCKET);

    signal(SIGINT, (void (*)(int))sigterm_handler);

    for (i=0; i<argc; i++)
    {
        /* Running with commands given as command line parameters. */
        stat = run_command(argv[i]);
        if ( stat == -1 ) 
        {
            /* The command ended up with an error, detach from the server and exit. */
            run_command(CMD_LEAVE);
            return stat;
        }
    }
    if ( i != 0 )
    {
        /* If the last command issued was not equal to CMD_QUIT, detach from the server. 
            If the command was equal to CMD_QUIT, the server already knows, let the client
            exit.
        */
        if ( cmdcmp(argv[i-1], CMD_QUIT) )
            run_command(CMD_LEAVE);

        return 0;
    }


    /* If we are here, the client should be run in the interactive mode. */

#ifdef HAVE_GTK
    if ( settings.viewfinder )
    {
        run_command(CMD_START);
        snprintf(str, MAX_CMD_LEN, "%s on", CMD_VF);
        run_command(str);
    }
#endif


#ifdef HAVE_READLINE
    using_history();
#endif


    while( 1 ) 
    {
#ifdef HAVE_READLINE
        s = readline("capture> ");
        if ( ! s )
        {
            stat = run_command(CMD_LEAVE);
            break;
        }
        if ( ! *s ) continue;
        add_history(s);
        stat = run_command(s);
#else
        printf("capture> ");
        fgets(str, MAX_CMD_LEN, stdin);
        if ( feof(stdin) ) 
        {
            stat = run_command(CMD_LEAVE);
            break;
        }
        if ( *str == '\n' ) continue;

        s = str;
        while (str-s < MAX_CMD_LEN && *s) s++;
        if ( s != str ) *(s-1) = 0;

        stat = run_command(str);
#endif
        if ( stat == 1 ) break;
    }

    close(SOCKET);

    if ( stat == -1 ) return stat;
    return 0;
}

