/* capture.c (c) 2004 Petr Danecek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <ctype.h>
#include "capture.h"
#include "ptp-utils.h"
#include "properties.h"

settings_t settings;

static int started = 0;

#define CHK_STARTED {if (!started) {msg_cmd_error("","start command expected"); return 0;} }
#undef RUN_N_CHK
#define RUN_N_CHK(x) { int stat = (x); if((stat)!=PTP_RC_OK) { \
      error("bad status on %s:%d - %04X\n", __FILE__, __LINE__, stat); \
      msg_cmd_error(cmd->command,"bad status: %04X\n", stat); \
      return 0; } }

cmd_t cmds[] = 
{
      { CMD_APERTURE,  PTP_DPC_CANON_Aperture, prop16_handler, aperture_vals, "Set the aperture"},
      { CMD_CAPTURE, 0, capture_handler,  NULL, "Capture image and save it to the given file name" },
#ifdef EXPERIMENTAL      
      { CMD_PHOTOGRAPH, 0, photograph_handler,  NULL, "Capture image but do not download it" },
#endif      
      { CMD_LEAVE,  0, leave_handler,  NULL, "Quit the client but let the server running"},
      { CMD_DEBUG, 0, debug_handler, debug_vals, "Toggle debug mode"},
      { CMD_EXPCOMP, PTP_DPC_CANON_ExpCompensation, prop8_handler, exposure_compensation_vals,"Exposure compensation"},
      { CMD_PHOTO_EFFECT, PTP_DPC_CANON_PhotoEffect, prop16_handler, photo_effect_vals, "Photo effect" },
      { CMD_FLASH,  PTP_DPC_CANON_FlashMode, prop8_handler, flash_vals, "Set flash mode" },
      { CMD_FOCUS_LOCK, 0, focuslock_handler, focus_vals, "Focus lock/unlock" },
      { CMD_FP,   PTP_DPC_CANON_FocusingPoint, prop16_handler, focuspoint_vals, "Sets the focusing point to center/auto" },
      { CMD_HELP, 0,  NULL,NULL, "More help on a commmand: \"help COMMAND\""},
      { CMD_ISO_SPEED, PTP_DPC_CANON_ISOSpeed, prop16_handler, iso_speed_vals, "ISO speed" },
      { CMD_MACRO,   PTP_DPC_CANON_MacroMode, prop8_handler, macro_vals, "Set the macro mode" }, 
      { CMD_LIGHT_METER, PTP_DPC_CANON_MeteringMode, prop8_handler, light_metering_vals, "Light metering mode" },
      { CMD_QUIT,   0, quit_handler,   NULL, "Quit both the server and client"},
      { CMD_SHUTTER, PTP_DPC_CANON_ShutterSpeed, prop16_handler, shutter_vals,  "Set the shutter speed" },
      { CMD_START,  0, start_handler,  NULL, "Initialize the camera"},
      { CMD_STATUS, 0, status_handler, NULL, "Get the server's status"},
      { CMD_TEST,   0, test_handler, NULL, "Debugging tests"},
      { CMD_TVAV, PTP_DPC_CANON_TvAvSetting, prop8_handler, tvav_vals,"TV/Av Setting - sets aperture/shutter priority"},
#ifdef HAVE_GTK
      { CMD_VF,       0,  viewfinder_handler, vf_vals,   "Turn the viewfinder window on/ff"},
      { CMD_SETVFOUTPUT, PTP_DPC_CANON_ViewfinderOutput, prop8_handler, vfout_vals, "Set the viewfinder output mode"},
      { CMD_VFSAVE,   0,  vfsave_handler, NULL,   "Save viewfinder images into files with a given prefix."},
      { CMD_VFALIGN,  0,  vfalign_handler, vfalign_vals, "Align the viewfinder window with the specified image."},
      { CMD_VFTRANS,  0,  vftrans_handler, vftrans_vals, "Set the viewfinder transparency in 'transparent' mode."},
      { CMD_BORDERS,  0,  borders_handler, borders_vals, "Draw borders around the viewfinder, in % of the height/width."},
#endif
      { CMD_WHITE_BAL, PTP_DPC_CANON_WhiteBalance, prop8_handler, white_balance_vals, "White balance"},
      { CMD_ZOOM,   PTP_DPC_CANON_Zoom, prop16_handler, zoom_vals, "Set zoom: 0..6 optical zoom, 7..10 digital zoom"},
      { CMD_QUAL,   PTP_DPC_CANON_ImageQuality, prop8_handler, qual_vals, "Set image quality"},
      { CMD_SIZE,   PTP_DPC_CANON_ImageSize, prop8_handler, size_vals, "Set image size"},
      { NULL, 0, NULL, NULL, NULL },
};




int quit_handler(cmd_t *cmd, char *args)
{
      msg_cmd_ok(CMD_QUIT);

      if ( started )
      {
#ifdef HAVE_GTK	   
	    if ( settings.viewfinder )
		  viewfinder_off(&settings);
#endif

	    debug("closing camera\n");
	    camera_close(&settings);
      }

      server_exit_nicely();
      return 0;
}


int start_handler(cmd_t *cmd, char *args)
{
      int busn, devn;
      int i;
      if ( started ) 
      {
	    debug("already started\n");
	    msg_cmd_ok(CMD_START);
	    return 0;
      }
      
      busn=devn=0;
      
      for (i=0; i<5; i++)
      {
	    if ( i ) 
	    {
		  debug("Trying again\n");
		  sleep(1);
	    }

	    settings.dev = find_device(busn,devn);

	    if ( ! settings.dev )
	    {
		  debug("no device matching bus=%d/dev=%d found\n", busn,devn);
		  continue;
	    }

	    find_endpoints(&settings);
	    if ( ! ptp_usb_init(&settings) ) 
	    {
		  debug("could not init ptp_usb\n");
	    }

	    sleep(1);

	    if ( (ptp_opensession(&(settings.params),1)) != PTP_RC_OK )
	    {
		  debug("could not open session\n");
		  ptp_usb_close(&settings);
		  continue;
	    }

	    break;
      }

      if (i == 5) 
      {
	    msg_cmd_error(CMD_START, "could not initialize camera");
	    return 0;
      }
      
      RUN_N_CHK( magic_code(&settings) );
      RUN_N_CHK( ptp_canon_startshootingmode(&(settings.params)) );
      /* The usb_checkevent_wait command is not critical and may fail(?) */
      usb_checkevent_wait(&settings); 
      ptp_checkevent(&settings,NULL);
      ptp_checkevent(&settings,NULL);
      RUN_N_CHK( set_prop16(&settings, PTP_DPC_CANON_D029, 3) );

      msg_cmd_ok(CMD_START);
      started = 1;
      return 0;
}

int photograph_handler(cmd_t *cmd, char *args)
{
      int result = 0;
      RUN_N_CHK( result=photograph(&settings) );

      msg_cmd_ok(cmd->command);
      return 0;
      /* return result; */
}

int status_handler(cmd_t *cmd, char *args)
{
      msg_cmd_ok(CMD_STATUS);
      return 0;
}


int debug_handler(cmd_t *cmd, char *args)
{
	int prop;


	prop = value_ok(cmd->params, args);

	if ( prop == -1 )
	{
		msg_cmd_error(cmd->command, "bad parameter \"%s\"", args);
		return 0;
	}

	/* enable or disable DEBUG mode */
	DEBUG = prop;
	
	msg_cmd_ok(cmd->command);
	return 0;
}



int prop16_handler(cmd_t *cmd, char *args)
{
    int prop;
    uint16_t value;
    PTPDevicePropDesc dpd;
    CHK_STARTED;

    memset(&dpd, 0, sizeof(dpd));

    if (getarg(args, NULL) == NULL) 
    {
        RUN_N_CHK(ptp_getdevicepropdesc(&settings.params, cmd->ptpcode, &dpd));
        msg_cmd_info(cmd->command , "Current value = \"%s\".", get_name_from_code(cmd->params, *(uint16_t*)dpd.CurrentValue));
        ptp_free_devicepropdesc(&dpd);
        return 0;
    }

    prop = value_ok(cmd->params, args);
    if ( prop == -1 )
    {
        msg_cmd_error(cmd->command, "bad parameter \"%s\"", args);
        return 0;
    }
    value = prop;

    /* the next two lines are needed to set the value */
    RUN_N_CHK(ptp_getdevicepropdesc(&settings.params,cmd->ptpcode,&dpd));
    ptp_free_devicepropdesc(&dpd);

    RUN_N_CHK(ptp_setdevicepropvalue(&settings.params, cmd->ptpcode, &value, dpd.DataType));
    RUN_N_CHK(ptp_getdevicepropdesc(&settings.params, cmd->ptpcode, &dpd));
    if (value != *(uint16_t*)dpd.CurrentValue) 
    {
        msg_cmd_error(cmd->command, "Set to \"%s\".",
                get_name_from_code(cmd->params, *(uint16_t*)dpd.CurrentValue));
    } 
    else
        msg_cmd_ok(cmd->command);

    ptp_free_devicepropdesc(&dpd);

    return 0;
}


int prop8_handler(cmd_t *cmd, char *args)
{
      int prop;
      PTPDevicePropDesc dpd;
      CHK_STARTED;

      memset(&dpd, 0, sizeof(dpd));

      if (getarg(args, NULL) == NULL) 
      {
          RUN_N_CHK(ptp_getdevicepropdesc(&settings.params, cmd->ptpcode, &dpd));
          msg_cmd_info(cmd->command , "Current value = \"%s\".", get_name_from_code(cmd->params, *(uint8_t*)dpd.CurrentValue));
          ptp_free_devicepropdesc(&dpd);
          return 0;
      }
      
      prop = value_ok(cmd->params, args);

      if ( prop == -1 )
      {
	    msg_cmd_error(cmd->command, "bad parameter \"%s\"", args);
	    return 0;
      }

      RUN_N_CHK( set_prop8(&settings, cmd->ptpcode, prop) );

      msg_cmd_ok(cmd->command);
      return 0;
}


int capture_handler(cmd_t *cmd, char *args)
{
      int n;
      char buf[MAX_CMD_LEN];
      CHK_STARTED;
      
      args = getarg(args, &n);
      
      if ( ! args )
      {
	    msg_cmd_error(cmd->command, "missing file name");
	    return 0;
      }
      
      if ( n+1 > MAX_CMD_LEN ) n = MAX_CMD_LEN-1;

      memcpy(buf, args, n);
      buf[n] = 0;

#ifdef HAVE_GTK
      if ( settings.viewfinder )
      {
	    ptp_canon_viewfinderoff(&(settings.params));
	    /* viewfinder_off(&settings); */
	    settings.viewfinderwait = 1; /* suspend only */

	    n = capture(&settings, buf);

	    ptp_canon_viewfinderon(&(settings.params));
	    /* viewfinder_on(&settings); */
	    settings.viewfinderwait = 0; /* unsuspend */
      }
      else
#endif
	    n = capture(&settings, buf);
	    
      if ( n != PTP_RC_OK )
	    msg_cmd_error(cmd->command, "bad status %04X\n", n);
      else
	    msg_cmd_ok(cmd->command);
      return 0;
}


int leave_handler(cmd_t *cmd, char *args)
{
      msg_cmd_ok(CMD_LEAVE);
      return 1;
}


#ifdef HAVE_GTK

int vfalign_handler(cmd_t *cmd, char *args)
{
      char *f;
      int n, align;
      CHK_STARTED;
     
      f = NULL;
     
      align = value1_ok(vfalign_vals, args, &n);
      if ( align == -1 )
      {
	    msg_cmd_error(cmd->command, "bad parameter \"%s\" (see \"help COMMAND\")", args);
	    return 0;
      }
      
      f = getarg(args+n, &n);

      /* no filename specified ? and you're not asking to switch it off ? */
      if ( ( ! f )  && ( align ) )
      {
	    msg_cmd_error(cmd->command, "missing filename");
	    return 0;
      }
      
      /* ok.. there's a file, then update */
      if ( f )
      {
	    /* would it be possible to use only alignfname to alert when the
	     * file changes and the viewfinder stuff will update
	     * when alignfname isn't NULL anymore. And over there.. after
	     * loading it would be reset to NULL */
	    settings.alignfname = strdup(f);
      }

      settings.alignview  = align;

      if ( ! settings.viewfinder ) viewfinder_on(&settings);

      msg_cmd_ok(cmd->command);
      return 0;
}



int borders_handler(cmd_t *cmd, char *args)
{
      char *f;
      int n, borders;
      CHK_STARTED;
     
      f = NULL;
     
      borders = value1_ok(borders_vals, args, &n);
      if ( borders == -1 )
      {
	    msg_cmd_error(cmd->command, "bad parameter \"%s\" (see \"help COMMAND\")", args);
	    return 0;
      }
      
      settings.bordersview  = borders;

      msg_cmd_ok(cmd->command);
      return 0;
}


int vftrans_handler(cmd_t *cmd, char *args)
{
      int n, opacity;
      CHK_STARTED;
     
      opacity = value1_ok(vftrans_vals, args, &n);
      if ( opacity == -1 )
      {
	    msg_cmd_error(cmd->command, "bad parameter \"%s\" (see \"help COMMAND\")", args);
	    return 0;
      }
      
      settings.viewfinder_opacity = opacity;

      msg_cmd_ok(cmd->command);
      return 0;
}


int viewfinder_handler(cmd_t *cmd, char *args)
{
      int vf;
      CHK_STARTED;
      
      vf = value_ok(cmd->params, args);

      switch (vf)
      {
	    case 2 : 
      /*  
		      viewfinder_off(&settings);
		      viewfinder_on(&settings);
      */
		  
		  /* FIXME: this code does not work well:
       * maybe libptp2-1.0.1 fixed this issue ?
       * seems to works well now
       */
		  RUN_N_CHK( ptp_canon_reflectchanges(&(settings.params), 7) );
		  RUN_N_CHK( getchanges(&settings) );

		  break;
	    case 1 : 
		  viewfinder_on(&settings); 
		  break;
	    case 0 : 
		  viewfinder_off(&settings); 
		  break;
	    default:
		  msg_cmd_error(cmd->command, "bad parameter \"%s\"", args);
		  return 0;
      }

      msg_cmd_ok(cmd->command);
      return 0;
}


int vfsave_handler(cmd_t *cmd, char *args)
{
    int n;
    
    args = getarg(args, &n);
    
    if ( ! args )
    {
        msg_cmd_error(cmd->command, "missing file name prefix to start capturing or \"off\" to stop\n");
        return 0;
    }

    if ( !strcasecmp(args, "off") ) 
    {
        free(settings.vfsave_file);
        settings.vfsave_file = NULL;
    }
    else
    {
        settings.vfsave_file = strdup(args);
        settings.vfsave_idx  = 0;
    }

    msg_cmd_ok(cmd->command);
    return 0;
}

#endif


int focuslock_handler(cmd_t *cmd, char *args)
{
      int lock;
      CHK_STARTED;

      lock = value_ok(cmd->params, args);

      if ( lock == -1 )
      {
	    msg_cmd_error(cmd->command, "bad parameter \"%s\"", args);
	    return 0;
      }

      if ( lock )
	    lock_focus(&settings);
      else
	    unlock_focus(&settings);
      
      msg_cmd_ok(cmd->command);
      return 0;
}


int test_handler(cmd_t *cmd, char *args)
{
      int stat;

      debug("Test handler called.\n");
      /* msg_cmd_ok(cmd->command); */
      msg_cmd_info(cmd->command,"info issued");
      return 0;
      
      debug("test_handler: trying to end shooting mode and close the session\n");
      
      if ( (stat=ptp_canon_endshootingmode(&(settings.params))) != PTP_RC_OK )
	    error("bad status: endshootingmode returned %04X\n", stat);
      
      stat = usb_checkevent_wait(&settings);
	    
      ptp_checkevent(&settings, NULL);
      ptp_checkevent(&settings, NULL);

      if ( stat != PTP_RC_OK )
	    usb_checkevent_wait(&settings);
	    
      if ( (stat=ptp_closesession(&(settings.params))) != PTP_RC_OK )
	    error("bad status: ptp_closesession returned %04X\n", stat);

      debug("test_handler: sleeping for 10sec\n");
      sleep(10);
      debug("test_handler: woked up\n");

      debug("test_handler: trying to startshooting mode again\n");
      RUN_N_CHK( ptp_canon_startshootingmode(&(settings.params)) );
      RUN_N_CHK( usb_checkevent_wait(&settings) );
      ptp_checkevent(&settings,NULL);
      ptp_checkevent(&settings,NULL);
      RUN_N_CHK( set_prop16(&settings, PTP_DPC_CANON_D029, 3) );

      msg_cmd_ok(cmd->command);
      return 0;
}


