/* capture.c (c) 2004 Petr Danecek
 * Based on work of Mariusz Woloszyn and Nikolai Kopanygin.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "ptp-utils.h"
#include "capture.h"

#ifndef USB_CLASS_PTP
#define USB_CLASS_PTP 6
#endif

#define MAX_SMALLREAD_SIZE 256
#define MAX_BULK_SIZE 131072


void usb_debug_event (PTPContainer *ev)
{
    debug("usb event: ");
    if ( ! ev )
    {
	debug("no event!\n");
	return;
    }

    debug(" code=");
    switch (ev->Code)
    {
	case PTP_EC_Undefined :		debug("undefined"); break;
	case PTP_EC_CancelTransaction :	debug("cancel transaction"); break;
	case PTP_EC_ObjectAdded :		debug("object added"); break;
	case PTP_EC_ObjectRemoved :		debug("object removed"); break;
	case PTP_EC_StoreAdded :		debug("store added"); break;
	case PTP_EC_StoreRemoved :		debug("store removed"); break;
	case PTP_EC_DevicePropChanged :	debug("device prop changed"); break;
	case PTP_EC_ObjectInfoChanged :	debug("object info changed"); break;
	case PTP_EC_DeviceInfoChanged :	debug("device info changed"); break;
	case PTP_EC_RequestObjectTransfer:	debug("request object transfer"); break;
	case PTP_EC_StoreFull :		debug("store full"); break;
	case PTP_EC_DeviceReset :		debug("device reset"); break;
	case PTP_EC_StorageInfoChanged :	debug("storage info changed"); break;
	case PTP_EC_CaptureComplete :	debug("capture complete"); break;
	case PTP_EC_UnreportedStatus :	debug("unreported status"); break;
	case PTP_EC_CANON_DeviceInfoChanged :     debug("c devinfo changed"); break;
	case PTP_EC_CANON_RequestObjectTransfer : debug("c reqobjecttransfer"); break;
	case PTP_EC_CANON_CameraModeChanged :     debug("c cameramodechanged"); break;
	default : debug("%04X", ev->Code); break;
    }
    debug("  sid=%04X", ev->SessionID);
    debug("  tid=%04X", ev->Transaction_ID);
    debug("  p1=%04X", ev->Param1);
    debug("  p2=%04X", ev->Param2);
    debug("  p3=%04X", ev->Param3);
    debug("\n");
}

void ptp_debug_event (int isevent, PTPUSBEventContainer *event)
{
    debug("ptp event: ");
    if ( ! isevent ) 
    {
	debug("no event!\n");
	return;
    }

    debug(" type=");
    switch (event->type)
    {
	case PTP_USB_CONTAINER_UNDEFINED:	debug("undefined"); break;
	case PTP_USB_CONTAINER_COMMAND:	debug("command"); break;
	case PTP_USB_CONTAINER_DATA:	debug("data"); break;
	case PTP_USB_CONTAINER_RESPONSE:	debug("response"); break;
	case PTP_USB_CONTAINER_EVENT:	debug("event"); break;
	default:	debug("%04X", event->type); break;
    }

    debug("  code=");
    switch (event->code)
    {
	case PTP_EC_CANON_CameraModeChanged:      debug("CameraModeChanged"); break;
	case PTP_EC_StorageInfoChanged:	      debug("StorageInfoChanged"); break;
	case PTP_EC_CANON_DeviceInfoChanged:      debug("DeviceInfoChanged"); break;
	case PTP_EC_CANON_RequestObjectTransfer:  debug("RequestObjectTransfer"); break;
	case PTP_EC_CaptureComplete:	      debug("CaptureCompleted"); break;
	default: debug("%04X", event->code); break;
    }
    debug("  len=%d", event->length);
    debug("  tid=%04X", event->trans_id);
    debug("  p1=%04X", event->param1);
    debug("  p2=%04X", event->param2);
    debug("  p3=%04X\n", event->param3);
}



static short ptp_check_int_func(unsigned char *bytes, unsigned int size, void *data)
{
    int count=100;
    int result=0;
    PTP_USB *ptp_usb=(PTP_USB *)data;
    while( !result && count ) 
    {
	result=usb_bulk_read(ptp_usb->handle, ptp_usb->intep, (char*)bytes, size,MAX_SMALLREAD_SIZE);
	count--;
    }
    if (result >= 0) 
    {

	do 
	{
	    bytes += result;
	    size  -= result;
	    if (size>0)
		result=usb_bulk_read(ptp_usb->handle, ptp_usb->intep, (char*)bytes, size,MAX_SMALLREAD_SIZE);
	} while ((result>0) && (size>0));

	return (PTP_RC_OK);
    } 

    debug("ptp_check_int_func usb_bulk_read: %s\n", strerror(errno));
    return PTP_ERROR_IO;
}



static short ptp_check_int_fast_func(unsigned char *bytes, unsigned int size, void *data)
{
    int result;
    PTP_USB *ptp_usb=(PTP_USB *)data;

    result=usb_bulk_read(ptp_usb->handle, ptp_usb->intep, (char*)bytes, size,MAX_SMALLREAD_SIZE);
    if ( ! result )
	result=usb_bulk_read(ptp_usb->handle, ptp_usb->intep, (char*)bytes, size,MAX_SMALLREAD_SIZE);
    if (result >= 0)
    {
	do 
	{
	    bytes+=result;
	    size-=result;
	    if (size>0)
		result=usb_bulk_read(ptp_usb->handle, ptp_usb->intep, (char*)bytes, size,MAX_SMALLREAD_SIZE);
	} 
	while ((result>0) && (size>0));

	return (PTP_RC_OK);
    } 

    debug("ptp_check_int_fast_func usb_bulk_read: %s\n", strerror(errno));
    return PTP_ERROR_IO;
}


static short ptp_read_func (unsigned char *bytes, unsigned int size, void *data)
{
    int result;
    PTP_USB *ptp_usb=(PTP_USB *)data;

    result=usb_bulk_read(ptp_usb->handle, ptp_usb->inep, (char*)bytes, size,MAX_BULK_SIZE);
    if ( ! result )
	result=usb_bulk_read(ptp_usb->handle, ptp_usb->inep, (char*)bytes, size,MAX_BULK_SIZE);
    if (result >= 0)
	return (PTP_RC_OK);
    else 
    {
	debug("ptp_read_func usb_bulk_read: %s\n", strerror(errno));
	return PTP_ERROR_IO;
    }
}


static short ptp_write_func (unsigned char *bytes, unsigned int size, void *data)
{
    int result;
    PTP_USB *ptp_usb=(PTP_USB *)data;

    result=usb_bulk_write(ptp_usb->handle,ptp_usb->outep,(char*)bytes,size,MAX_BULK_SIZE);
    if (result >= 0)
	return (PTP_RC_OK);

    debug("ptp_write_func usb_bulk_write: %s\n", strerror(errno));
    return PTP_ERROR_IO;
}



struct usb_device* find_device (int busn, int devn)
{
    struct usb_bus *bus;
    struct usb_device *dev;

    usb_init();
    usb_find_busses();
    usb_find_devices();
    bus = usb_get_busses();

    for (; bus; bus = bus->next)
	for (dev = bus->devices; dev; dev = dev->next)
	    if (dev->config->interface->altsetting->bInterfaceClass==USB_CLASS_PTP)
		if (dev->descriptor.bDeviceClass!=USB_CLASS_HUB)
		{
		    int curbusn, curdevn;

		    curbusn=strtol(bus->dirname,NULL,10);
		    curdevn=strtol(dev->filename,NULL,10);

		    if (devn==0) 
		    {
			if (busn==0) return dev;
			if (curbusn==busn) return dev;
		    } 
		    else 
		    {
			if ((busn==0)&&(curdevn==devn)) return dev;
			if ((curbusn==busn)&&(curdevn==devn)) return dev;
		    }
		}
    return NULL;
}


void find_endpoints(settings_t *settings)
{
    int i,n;
    struct usb_endpoint_descriptor *ep;

    ep = settings->dev->config->interface->altsetting->endpoint;
    n  = settings->dev->config->interface->altsetting->bNumEndpoints;

    for (i=0; i<n; i++) 
    {
	if (ep[i].bmAttributes==USB_ENDPOINT_TYPE_BULK) 
	{
	    if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)== USB_ENDPOINT_DIR_MASK)
		settings->ptp_usb.inep  = ep[i].bEndpointAddress;

	    if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==0)
		settings->ptp_usb.outep = ep[i].bEndpointAddress;
	} 
	else
	{
	    if (ep[i].bmAttributes==USB_ENDPOINT_TYPE_INTERRUPT)
	    {
		if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)== USB_ENDPOINT_DIR_MASK)
		    settings->ptp_usb.intep = ep[i].bEndpointAddress;
	    }
	}
    }
}


void Error(void *data, const char *format, va_list args)
{
    debug(format,args);
    debug("\n");
}

void Debug(void *data, const char *format, va_list args)
{
    debug(format,args);
    debug("\n");
}


int ptp_usb_init(settings_t *settings)
{
    usb_dev_handle *device_handle;

    settings->params.write_func	  = ptp_write_func;
    settings->params.read_func	  = ptp_read_func;
    settings->params.check_int_func	  = ptp_check_int_func;
    settings->params.check_int_fast_func	= ptp_check_int_fast_func;
    settings->params.error_func	  = Error;
    settings->params.debug_func	  = Debug;
    settings->params.sendreq_func	  = ptp_usb_sendreq;
    settings->params.senddata_func	  = ptp_usb_senddata;
    settings->params.getresp_func	  = ptp_usb_getresp;
    settings->params.getdata_func	  = ptp_usb_getdata;
    settings->params.data		  = &(settings->ptp_usb);
    settings->params.transaction_id	  = 0;
    settings->params.byteorder	  = PTP_DL_LE;

    device_handle = usb_open(settings->dev);

    if ( device_handle )
    {
	settings->ptp_usb.handle = device_handle;
	usb_claim_interface(
		device_handle, 
		settings->dev->config->interface->altsetting->bInterfaceNumber
		);

	return 1;
    }

    return 0;
}


void ptp_usb_close(settings_t *settings)
{
    usb_release_interface(settings->ptp_usb.handle,
	    settings->dev->config->interface->altsetting->bInterfaceNumber);
}


int get_prop16 (settings_t *settings, uint16_t prop, uint16_t *val)
{
    uint16_t *p = NULL;

    uint32_t stat = ptp_getdevicepropvalue(&(settings->params), prop, (void **)&p, PTP_DTC_UINT16);

    if ( p && *p )
    {
	*val = *p;
	free(p);
    }

    debug("getting %04X = %04x \t (0x%04X)\n", prop, *val, stat);
    return stat;
}

int get_prop32 (settings_t *settings, uint16_t prop, uint32_t *val)
{
    uint32_t *p = NULL;

    uint32_t stat = ptp_getdevicepropvalue(&(settings->params), prop, (void **)&p, PTP_DTC_UINT32);

    if ( p && *p )
    {
	*val = *p;
	free(p);
    }

    debug("getting %04X = %04x \t (0x%04X)\n", prop, *val, stat);
    return stat;
}

int set_prop8(settings_t *settings, uint16_t prop, uint8_t val)
{
    uint32_t stat;
    debug("setting %04X to %04x  ", prop, val);
    stat = ptp_setdevicepropvalue(&(settings->params), prop, &val, PTP_DTC_UINT8);
    debug("\t (0x%04X)\n", stat);
    return stat;
}

int set_prop16(settings_t *settings, uint16_t prop, uint16_t val)
{
    uint32_t stat;
    debug("setting %04X to %04x  ", prop, val);
    stat = ptp_setdevicepropvalue(&(settings->params), prop, &val, PTP_DTC_UINT16);
    debug("\t (0x%04X)\n", stat);
    return stat;
}


int magic_code(settings_t *settings)
{
    uint32_t ui32;
    uint16_t ui16;

    /* RUN_N_CHK( get_prop16(settings, 0xd045, &ui16) ); */
    get_prop16(settings, 0xd045, &ui16);
    RUN_N_CHK( set_prop16(settings, 0xd045, 1) );
    RUN_N_CHK( get_prop32(settings, 0xd02e, &ui32) );
    RUN_N_CHK( get_prop32(settings,0xd02f, &ui32) );

    RUN_N_CHK( ptp_getdeviceinfo(&(settings->params), &(settings->params.deviceinfo)) );
    RUN_N_CHK( ptp_getdeviceinfo(&(settings->params), &(settings->params.deviceinfo)) );

    RUN_N_CHK( get_prop32(settings, 0xd02e, &ui32) );
    RUN_N_CHK( get_prop32(settings, 0xd02f, &ui32) );

    RUN_N_CHK( ptp_getdeviceinfo(&(settings->params), &(settings->params.deviceinfo)) );

    RUN_N_CHK( get_prop16(settings, 0xd045, &ui16) );
    RUN_N_CHK( set_prop16(settings, 0xd045, 4) );

    return PTP_RC_OK;
}


int ptp_checkevent(settings_t *settings, PTPUSBEventContainer *fill_event)
{
    int isevent;
    PTPUSBEventContainer event, *ev = (fill_event) ? fill_event : &event;


    if ( ptp_canon_checkevent(&(settings->params),ev,&isevent) != PTP_RC_OK ) return 0;
    if ( ! isevent ) return 0;
    if ( ! ev->code && ev->type == PTP_USB_CONTAINER_UNDEFINED ) return 0;

    ptp_debug_event(isevent, ev);
    return 1;
}


int usb_checkevent_wait(settings_t *settings)
{
    int stat;
    PTPContainer evc;

    stat = ptp_usb_event_wait(&(settings->params), &evc);

    usb_debug_event(&evc);

    return stat;
}

int photograph(settings_t *settings)
{
    /* take a photograph.  don't bother bringing it back over USB. */
    PTPUSBEventContainer event;
    int i, stat, tries=50;

    /* trigger the camera to do the actual capture. */

    RUN_N_CHK( ptp_initiatecapture(&(settings->params), 0x00000000, PTP_OFC_JP2) );
    stat = usb_checkevent_wait(settings);
    /* wait for the camera to come back into a sensible state (?) */
    for (i=0; i<tries; i++)
    {
      if ( i ) usleep(USLEEP);
      if ( ! ptp_checkevent(settings, &event) ) continue;

      if (  event.type != PTP_USB_CONTAINER_EVENT ) continue; 

      ptp_checkevent(settings, NULL);
      ptp_checkevent(settings, NULL);

      break;
    }

    if ( stat != PTP_RC_OK )
      stat = usb_checkevent_wait(settings);

    if ( i == tries ) 
    {
      error("... capture timed out ...\n");
      return PTP_ERROR_IO;
    }

    return PTP_RC_OK;
}

int capture(settings_t *settings, char *fname)
{
    PTPUSBEventContainer event;
    int i,j, nblocks, bsize=5000, stat, tries=50;
    unsigned int readnum;
    uint32_t tail, pos, handle = 0, size = 0, dummy = 0;
    FILE *fh;
    unsigned char *image = NULL;

    /* trigger the camera to do the actual capture. */

    RUN_N_CHK( ptp_canon_initiatecaptureinmemory(&(settings->params)) );
    stat = usb_checkevent_wait(settings);
    /* wait for the camera to come back into a sensible state (?) */
    for (i=0; i<tries; i++)
    {
	if ( i ) usleep(USLEEP);
	if ( ! ptp_checkevent(settings, &event) ) continue;

	if (  event.type != PTP_USB_CONTAINER_EVENT || 
		event.code != PTP_EC_CANON_RequestObjectTransfer ) continue;

	handle = event.param1;

	ptp_checkevent(settings, NULL);
	ptp_checkevent(settings, NULL);

	break;
    }

    if ( stat != PTP_RC_OK )
	stat = usb_checkevent_wait(settings);

    if ( i == tries ) 
    {
	error("... capture timed out ...\n");
	return PTP_ERROR_IO;
    }
    /* get size of the picture */
    stat = ptp_canon_getobjectsize(&(settings->params), handle, 0, &size, &dummy);
    if ( stat != PTP_RC_OK )
    {
	error("could not get object size of %04X - status=%04X. trying again...\n", handle,stat);
	usleep(USLEEP);
	stat = ptp_canon_getobjectsize(&(settings->params), handle, 0, &size, &dummy);
	if ( stat != PTP_RC_OK )
	{
	    error("did not get object size: %04X\n", stat);
	    return stat;
	}
    }

    debug("object size=%d dummy=%d\n",size,dummy);

    if ( ! size ) 
    {
	error("no size\n");
	return PTP_ERROR_IO;
    }
    /* prepare outfile for writing */

    fh = fopen(fname, "w");
    if ( ! fh ) 
    {
	error("%s: %s\n", fname, strerror(errno));
	return PTP_ERROR_IO;
    }

    nblocks = size/bsize;
    tail    = size%bsize;

    for (i=0; i<nblocks; i++)
    {
	/* 1 start, 2 middle, 3 end */
	pos   = (!i) ? 0 : ( (!tail && i==nblocks-1) ? 2:1 );


	/* FIXME: during a very long serie of captures, the camera gets into an
	 * undefined state. Places of failure:
	 *   - we dont get PTP_EC_CANON_RequestObjectTransfer      
	 *   - getpartialobject
	 * The first failure seems to be recoverable - next capture may run ok,
	 * but the second one is not. Let's try to get partial object more than once.
	 * the image appears to come off the camera in blocks, which wind up 
	 * in *image.
	 */

	for (j=0; j<10; j++)
	{
	    image = NULL;
	    stat  = ptp_canon_getpartialobject(&(settings->params), handle, i*bsize, 
		    bsize, pos, &image, &readnum);
	    if ( stat != PTP_RC_OK )
	    {
		error("could not ptp_canon_getpartialobject: %04X. trying again...\n",stat);
		debug("error on: size=%d i=%d, bsize=%d, pos=%d, readnum=%d\n", size,i,bsize,pos,readnum);
		usleep(USLEEP);
		continue;
	    }

	    break;
	}

	if ( j == 10 )
	{
	    if (image) 
	    {
		debug("FIXME: how can be image != NULL with this status\n");
		free(image);
	    }
	    error("ptp_canon_getpartialobject failure\n");
	    fclose(fh);
	    return PTP_ERROR_IO;
	}

	if ( ! image ) 
	{
	    error("image read error\n");
	    fclose(fh);
	    return PTP_ERROR_IO;
	}
	fwrite(image, 1, readnum, fh);
	free(image);
	image = NULL;
    }

    if ( tail )
    {
	pos   = (nblocks) ? 3 : 1;

	image = NULL;
	stat  = ptp_canon_getpartialobject(&(settings->params), handle, i*bsize, 
		bsize, pos, &image, &readnum);
	if ( stat != PTP_RC_OK )
	{
	    error("could not ptp_canon_getpartialobject: %04X\n",stat);
	    debug("size=%d i=%d, bsize=%d, pos=%d, readnum=%d\n", size,i,bsize,pos,readnum);

	    if ( image ) free(image);
	    fclose(fh);
	    return PTP_ERROR_IO;
	}

	if ( ! image ) 
	{
	    error("image read error\n");
	    fclose(fh);
	    return PTP_ERROR_IO;
	}

	fwrite(image, 1, readnum, fh);
    }

    if (image) free(image);

    fclose(fh);

    return PTP_RC_OK;
}


int camera_close(settings_t *settings)
{
    int stat;

    if ( (stat=ptp_canon_endshootingmode(&(settings->params))) != PTP_RC_OK )
	error("bad status: endshootingmode returned %04X\n", stat);

    stat = usb_checkevent_wait(settings);

    ptp_checkevent(settings, NULL);
    ptp_checkevent(settings, NULL);

    if ( stat != PTP_RC_OK )
	usb_checkevent_wait(settings);

    if ( (stat=ptp_closesession(&(settings->params))) != PTP_RC_OK )
	error("bad status: ptp_closesession returned %04X\n", stat);

    ptp_usb_close(settings);
    return 1;
}

int getchanges(settings_t *settings)
{
    uint16_t *props = NULL;
    uint32_t propnum;
    int i;

    RUN_N_CHK( ptp_canon_getchanges(&(settings->params), &props, &propnum) )
	for (i=0; i<propnum; i++) debug("changed: %04X\n", props[i]);

    if ( props ) free(props);

    return PTP_RC_OK;
}


int lock_focus(settings_t *settings)
{
    RUN_N_CHK( ptp_canon_focuslock(&(settings->params)) );
    RUN_N_CHK( getchanges(settings) );
    return PTP_RC_OK;
}


int unlock_focus(settings_t *settings)
{
    RUN_N_CHK( ptp_canon_focusunlock(&(settings->params)) );
    RUN_N_CHK( getchanges(settings) );
    return PTP_RC_OK;
}

