/* capture.c (c) 2004 Petr Danecek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include "capture.h"
#include "ptp-utils.h"

static struct sockaddr_un local, remote;
int DEBUG     = 0;
int SOCKET    = 0;


static int started;

void server_exit_nicely()
{
      debug("exiting\n");

      debug("deleting temporary files\n");

#ifdef HAVE_GTK
      if ( settings.tmpalign )
	    unlink(settings.tmpalign);
#endif

      if ( log_file )
	    fclose(log_file);

      usleep(USLEEP);

      if ( SOCKET ) 
      {
	    close(SOCKET);
	    unlink(local.sun_path);
      }
      exit(1);
}

void ssigterm_handler(int signal, struct sigcontext sc)
{
      if ( started )
      {
#ifdef HAVE_GTK	   
	    if ( settings.viewfinder )
		  viewfinder_off(&settings);
#endif

	    debug("closing camera\n");
	    camera_close(&settings);
      }

      server_exit_nicely();
}


void run_server(void)
{
      int s, len;
      socklen_t t;
      char str[MAX_CMD_LEN];
      int flags;
      int done;

      if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) 
      {
	    error("socket: %s\n", strerror(errno));
	    exit(1);
      }

      signal(SIGINT, (void (*)(int))ssigterm_handler);

      local.sun_family = AF_UNIX;
      strcpy(local.sun_path, SOCK_PATH);
      unlink(local.sun_path);
      len = strlen(local.sun_path) + sizeof(local.sun_family);
      if (bind(s, (struct sockaddr *)&local, len) == -1) 
      {
	    error("bind \"%s\": %s\n", SOCK_PATH, strerror(errno));
	    exit(1);
      }

      if (listen(s, 5) == -1) 
      {
	    error("listen: %s\n", strerror(errno));
	    exit(1);
      }

      fcntl(s, F_GETFL, flags);
      fcntl(s, F_SETFL, flags|O_NONBLOCK);


      while (1)
      {
	    debug("Waiting for a connection...\n");

	    t = sizeof(remote);
	    while ( (SOCKET = accept(s, (struct sockaddr *)&remote, &t)) == -1 )
	    {
		  if ( errno != EAGAIN )
		  {
			error("accept: %s\n", strerror(errno));
			exit(1);
		  }
#ifdef HAVE_GTK		  
		  if ( settings.gtk ) 
			gtk_main_iteration_do(0);
#endif

		  usleep(USLEEP);
	    }

	    fcntl(SOCKET, F_GETFL, flags);
	    fcntl(SOCKET, F_SETFL, flags|O_NONBLOCK);

	    debug("Server: Connected \"%s\" as %d.\n", SOCK_PATH, SOCKET);

	    done = 0;
	    while (! done)
	    {
		  while ( ! msg_get(str, MAX_CMD_LEN) ) 
		  {
#ifdef HAVE_GTK
			if ( settings.gtk ) 
			      gtk_main_iteration_do(0);
#endif                  
			usleep(USLEEP);
		  }
		  done = exec_cmd(str);
	    } 
	    usleep(USLEEP);
	    close(SOCKET);
      }
}


char *getarg(char *line, int *n)
{
      char  *p = line;
      char  *s;

      while ( p && *p && isspace(*p) ) p++;
      
      if ( ! p || ! *p ) return NULL;
      
      s = p;
      while ( p && *p && ! isspace(*p) ) p++;
      
      if ( p == s ) return NULL;

      if (n != NULL) *n = p-s;
      return s;
}

int cmdcmp(char *cmd1, char *cmd2)
{
      int n1,n2;
      char *arg1 = getarg(cmd1, &n1);
      char *arg2 = getarg(cmd2, &n2);

      if ( n1 != n2 ) return 1;
      return strncasecmp(arg1,arg2,n1);
}


int exec_cmd(char *arg)
{
      int n;
      cmd_t *c;

      arg = getarg(arg, &n);
      if ( ! arg ) 
      {
	    msg_send(CMD_UNKNOWN);
	    return 0;
      }

      c = cmds;
      while ( c->command )
      {
	    if ( ! cmdcmp(arg, c->command) ) 
	    {
		  if ( c->handler )
		  {
			arg = getarg(arg+n, &n);
			return c->handler(c, arg);
		  }
		  else
		  {
			error("FIXME: no c->handler!\n");
			exit(1);
		  }
	    }
	    c++;
      }

      msg_send(CMD_UNKNOWN);

      return 0;
}

