/* viewfinder.c (c) 2004 Petr Danecek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "capture.h"
#include "ptp-utils.h"
#ifdef HAVE_GTK

#include <stdio.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>

#ifdef HAVE_MAGICK
#include <magick/magick.h>
#endif /* HAVE_MAGICK */

typedef struct
{
	int width;
	int height;
    int rowstride;    /* rowstride is: no_pixels_wide*num_channels*num_bytes_per_channel.
                       * It would be different if we used some other type of pictures, but
                       * we limit the vfalign'ed images to 3 channels RGB, 8 bits per channel */
    guchar *data;     /* from systems.h: unsigned char. */
}
image_t;


static int refresh_rate  = 10;

static image_t align_img = {0,0,0,NULL};
static char lastalignfname[255] = "";
static char tmpalign[] = "/tmp/capture-aXXXXXX";

static GtkWidget *vfwin  = NULL;
static GtkWidget *gtkim  = NULL;
static GdkPixmap *pixmap = NULL;
static int viewfinder_handler_id;

#ifdef HAVE_MAGICK
static Image vfaligned_image;
static Image viewfinder_image;
#endif /* HAVE_MAGICK */


#ifdef HAVE_MAGICK
void set_image_data_from_memory(char *ptr, size_t len, image_t *im)
{
	Image *myimg;
	ImageInfo imginfo;
	memset(&imginfo, 0, sizeof(ImageInfo)); /* reset the thing */

	imginfo->depth = 8;
	imginfo->colorspace = RGBColorspace; /* enum ColorspaceType */

	im->width    = -1;
	im->height   = -1;
	im->data     = NULL;

	myimg = BlobToImage( );

}
#endif /* HAVE_MAGICK */

void set_image_data_from_file(char *fname, image_t *im)
{
    GtkWidget *imw;
    GdkPixbuf *pixbuf;
    guchar *buf;

    im->width  = -1;
    im->height = -1;
    im->data   = NULL;

    imw = gtk_image_new_from_file(fname);
    if ( GTK_IMAGE(imw)->storage_type != GTK_IMAGE_PIXBUF )
    {
        debug("not a GTK_IMAGE_PIXBUF!\n");
        gtk_widget_destroy(imw);
        return;
    }
    if ( GTK_IMAGE(imw)->storage_type == GTK_IMAGE_EMPTY )
    {
        debug("GTK_IMAGE_EMPTY!\n");
        gtk_widget_destroy(imw);
        return;
    }
    pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(imw));

    if ( ! pixbuf )
    {
        error("no pixbuf!\n");
        gtk_widget_destroy(imw);
        return;
    }
    if ( gdk_pixbuf_get_colorspace(pixbuf) != GDK_COLORSPACE_RGB )
    {
        error("not rgb!\n");
        gtk_widget_destroy(imw);
        return;
    }
    if ( gdk_pixbuf_get_bits_per_sample (pixbuf) != 8 )
    {
        error("not 8 bits per sample!\n");
        gtk_widget_destroy(imw);
        return;
    }
    if ( gdk_pixbuf_get_n_channels(pixbuf) != 3 )
    {
        error("not 3 channels!\n");
        gtk_widget_destroy(imw);
        return;
    }

    buf = gdk_pixbuf_get_pixels(pixbuf);
    im->width     = gdk_pixbuf_get_width(pixbuf);
    im->height    = gdk_pixbuf_get_height(pixbuf);
    im->rowstride = gdk_pixbuf_get_rowstride(pixbuf);
    im->data      = (guchar *) malloc(sizeof(guchar)*im->rowstride*im->height);
    memcpy(im->data, buf, sizeof(guchar)*im->rowstride*im->height);

    gtk_widget_destroy(imw);
}



/**
 * image_t *a = image from vfalign command
 * image_t *b = image from viewfinder => the buffer we use to display.
 */
void combine_images(settings_t *settings, image_t *a, image_t *b)
{
        int i,j;
	int k;
	float opacity_viewfinder = (float) settings->viewfinder_opacity / 100;
	float opacity_still      = 1 - opacity_viewfinder;
    int w = a->width;
    int h = a->height;
    int r = a->rowstride;
	guchar *rd, *wr, res;
    guchar white[] = {0xff, 0xff, 0xff};
        
    switch (settings->alignview)
    {
        case 6:
            /* do true transparency */
            wr = b->data;
            rd = a->data;
            for ( j = 0 ; j < (r*h) ; j++ , wr++, rd++ )
            {
                res = ( *rd * opacity_still + *wr * opacity_viewfinder);
                memcpy(wr, &res, 1);
            }
            break;
        case 5:
            /* do pseudo-transparency, interpolate one pixel at a time */
            k = 0;
            for ( j = 0 ; j < (w*h) ; j++ )
            {
                if (!(j%w))
                    k = (k ? 0 : 1);     /* end of line, toggle line index */
                if ( k )
                {
                    /* one pixel, three channels (RGB) at a time */
                    memcpy(&(b->data[3*j]), &(a->data[3*j]), 3);
                    k = 0;
                }
                else
                    k = 1;
            }
            break;
        case 4:
            for (j=0; j<h; j++)
            {
                if ( (j/20)%2 )
                {
                    for (i=0; i<w; i+=40)
                        memcpy(&(b->data[3*(w*j+i)]), &(a->data[3*(w*j+i)]), 3*20);
                }
                else
                {
                    for (i=20; i<w; i+=40)
                        memcpy(&(b->data[3*(w*j+i)]), &(a->data[3*(w*j+i)]), 3*20);
                }
            }
            break;

        case 3:
            for (j=0; j<h; j++)
                for (i=0; i<w; i+=40)
                    memcpy(&(b->data[3*(w*j+i)]), &(a->data[3*(w*j+i)]), 3*20);
            break;

        case 2:
            for (j=0; j<h; j++)
            {
                if ( (j/20)%2 ) continue;
                memcpy(&(b->data[3*w*j]), &(a->data[3*w*j]), 3*w);
            }
            break;

        default:
            for (j=0; j<h/2; j++)
                memcpy(&(b->data[3*w*j]), &(a->data[3*w*j]), 3*w/2);
            for (j=h/2; j<h; j++)
                memcpy(&(b->data[3*(w*j+w/2)]), &(a->data[3*(w*j+w/2)]), 3*w/2);
    }

    /* draw a white rectangle */
    if (settings->bordersview)
    {
        j = settings->bordersview; /* percent of the width */

        /* Loops to draw the borders */

        /* draw a first line
           3*h = one line       h*j/200 = percentage of the lines to first skip
           w*j/200 = number of pixels to skip before to start drawing the line
         */
        wr = b->data;

        wr += r*(h*j/200); /* skip j/2 % of the lines */
        wr += 3*w*j/200; /* advance j/2 % of the width */

        for(i = 0; i < w*(100-j)/100; i++)
        {
            memcpy(wr, white, 3);
            wr += 3;
        }

        /* draw a second line                   
         */
        wr = b->data;

        wr += r*(h*(100-j/2)/100); /* skip 100-j/2 % of the lines */
        wr += 3*w*j/200; /* advance j/2 % of the width */

        /*printf("H: %d W: %d R: %d\n", h, w, r);
          printf("My value: %d  (h*(100-j/2)/100) H: %d J: %d\n", (h*(100-j/2)/100), h, j);*/
        for(i = 0; i < w*(100-j)/100; i++)
        {
            memcpy(wr, white, 3);
            wr += 3;
        }

        /* draw vertical segments */
        wr = b->data + r*(h*j/200);  /* first row */
        rd = b->data + r*(h*(100-j/2)/100);  /* stop when you pass that point */

        for( ; wr < rd ; wr += r) 
        { /* skip a row each time */
            /* left dot/segment */
            memcpy(&wr[3*w*j/200], white, 3);
            /* right dot/segment */
            memcpy(&wr[3*w*(100-j/2)/100], white, 3);
        }

    }
}



int viewfinder_refresh_handler(gpointer data)
{
    settings_t *settings = (settings_t *) data;

    uint32_t size;
    unsigned char *image = NULL;
    char fname[] = "/tmp/captureXXXXXX"; 
    int  fd, fd1;
    FILE *jpgfile;
    char cmd[MAX_CMD_LEN];
    image_t snap = {0,0,0,NULL};
    

    debug("call viewfinder_refresh_handler\n");

    /* don't update viewfinder while we capture an image */
    if ( settings->viewfinderwait )
        return 1;

    ptp_canon_getviewfinderimage(&(settings->params), &image, &size);

    fd = mkstemp(fname);
    close(fd);

    jpgfile = fopen(fname, "w+");
    fwrite(image, 1, size, jpgfile);
    fclose(jpgfile);
    free(image);
    image = NULL;

    set_image_data_from_file(fname, &snap);
    if ( ! snap.data ) 
    {
        viewfinder_off(settings);
        return 1;
    }

    /* we enabled vfalign ? */
    if ( settings->alignview )
    {
        /* same pic. we last loaded ? */
        /* both defined ? */
        if ( settings->alignfname && *lastalignfname )
        {
            if ( !strcmp(settings->alignfname, lastalignfname) )
            {
                /* it's already in memory! */
                free(settings->alignfname);
                settings->alignfname = NULL;
            }
        }

        /* load new file in buffer if we have a file name */
        /*	      if ( ! align_img.data || settings->alignfname ) {*/
        if ( settings->alignfname ) 
        {
            /* load from file */
            set_image_data_from_file(settings->alignfname, &align_img);

            /* need to resize vfaligned image ? */
            if ( align_img.width != snap.width || align_img.height != snap.height )
            {
                debug("align image must be scaled to %dx%d\n", snap.width, snap.height);

                fd1 = mkstemp(tmpalign);
                close(fd1);

                debug("tmpalign %s\n", tmpalign);

                snprintf(cmd,250,"convert -size %dx%d \"%s\" -resize %dx%d \"%s\"",
                    snap.width,snap.height,settings->alignfname,snap.width,snap.height,tmpalign);

                system(cmd);

                settings->tmpalign = tmpalign;

                free(align_img.data);
                set_image_data_from_file(settings->tmpalign, &align_img);
            }

            strcpy(lastalignfname, settings->alignfname);
            free(settings->alignfname);
            settings->alignfname = NULL;

        }

        combine_images(settings, &align_img, &snap);
    } /* end: if vfalign enabled */

    if ( ! gtkim )
    {
        gtkim = gtk_drawing_area_new();
        gtk_drawing_area_size(GTK_DRAWING_AREA(gtkim), snap.width, snap.height); 
        gtk_container_add (GTK_CONTAINER(vfwin), gtkim);
        gtk_widget_show_all(vfwin);
    }

    if ( ! pixmap )
    {
        debug("XX: new pixmap\n");
        pixmap = gdk_pixmap_new(vfwin->window, snap.width, snap.height, -1);
    }

    gdk_draw_rgb_image(pixmap, gtkim->style->fg_gc[GTK_WIDGET_STATE(gtkim)], 
            0,0, snap.width, snap.height, GDK_RGB_DITHER_NONE, snap.data, snap.rowstride);
    gdk_draw_pixmap(gtkim->window, gtkim->style->fg_gc[GTK_WIDGET_STATE (gtkim)], 
            pixmap, 0, 0, 0, 0, snap.width, snap.height);

    free(snap.data);
    if ( settings->vfsave_file )
    {
        /* If set, the viewfinder image will not be deleted, but it will be moved to
            a file specified by the given template settings.vfsave_file.
        */
        snprintf(cmd,MAX_CMD_LEN,"mv \"%s\" \"%s%06d.jpg\"",fname,settings->vfsave_file,settings->vfsave_idx++);
        system(cmd);
    }
    else
        unlink(fname);

    return 1;
}




void viewfinder_on(settings_t *settings)
{
      int argc;
      if ( vfwin ) return;

      if ( ! settings->gtk ) 
      {
	    char **argv = malloc(2*sizeof(char *));
	    argv[0]     = "viewfinder";
	    argv[1]     = "";
	    argc    = 1;
	    
	    if ( ! gtk_init_check(&argc,&argv) )
	    {
		  debug("could not initialize gtk!\n");
		  return;
	    }
	    settings->gtk = 1;
      }

      vfwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);

      ptp_canon_viewfinderon(&(settings->params));

      viewfinder_refresh_handler(settings);

      /* you: why don't you call viewfinder_refresh_handler periodically ? */
      viewfinder_handler_id = g_timeout_add(refresh_rate, viewfinder_refresh_handler, settings);
      g_main_context_wakeup(NULL);
      /* it has worked some time.. so maybe the problem is deeper */

      settings->viewfinder  = 1;
}


void viewfinder_off(settings_t *settings)
{
      if ( ! vfwin ) return;
      
      ptp_canon_viewfinderoff(&(settings->params));

      g_source_remove(viewfinder_handler_id);
      settings->viewfinder  = 0;
      viewfinder_handler_id = 0;

      gtk_widget_destroy(vfwin);

      gtkim = NULL;
      vfwin = NULL;
}


#endif /* HAVE_GTK */

